import logging
import os
import re

import sau
import sau.helpers

EIX_UPDATE_PATH='/usr/bin/eix-update'
RC_SERVICE_PATH='/sbin/rc-service'
EMERGE_PATH='/usr/bin/emerge'
EQUERY_PATH='/usr/bin/equery'

# parsing output from eix -Ttnc
package_re = re.compile('^\[([^\]])\] ([^ ]*) \((.*)\): .*$')
# parsing version information from substrings of the above
slot_re = re.compile('^(\(~\))?([^\(]+)(\([^\)]+\))$')

def identify_service_from_bin(exe):
    log = logging.getLogger(sau.LOGNAME)

    init_script_re = re.compile(r'/etc/init\.d/(.*)')

    cmd = [ EQUERY_PATH, '-Cq', 'b', exe ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if ret != 0:
        log.warning("searching for owner of {} failed:".format(exe))
        for line in out.splitlines():
            log.warning("stdout: {}".format(line))
        for line in err.splitlines():
            log.warning("stderr: {}".format(line))
        return None
    
    pkg = out.strip()
    cmd = [ EQUERY_PATH, '-Cq', 'f', pkg ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if ret != 0:
        log.warning("listing files for package {} failed:".format(pkg))
        for line in out.splitlines():
            log.warning("stdout: {}".format(line))
        for line in err.splitlines():
            log.warning("stderr: {}".format(line))
        return None
    matches = set()
    for line in out.splitlines():
        match = re.match(init_script_re, line)
        if match:
            matches.add(match.group(1))
    if len(matches) < 1:
        log.warning('Could not find any init script in package {}'.format(pkg))
    elif len(matches) > 1:
        log.warning('Found multiple init script in package {}'.format(pkg))
    else:
        return matches.pop()
    return None
    

def restart_service(service):
    log = logging.getLogger(sau.LOGNAME)
    cmd = [ RC_SERVICE_PATH, service, 'restart' ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if ret != 0:
        log.warning("Restart of {} failed:".format(service))
        for line in out.splitlines():
            log.warning("stdout: {}".format(line))
        for line in err.splitlines():
            log.warning("stderr: {}".format(line))
    else:
        log.info("restarted service {}".format(service))

def system_upgrade():
    log = logging.getLogger(sau.LOGNAME)
    log.debug('Gentoo has no concept of system upgrade, ignoring...')
    return False

def _sync_portage():
    log = logging.getLogger(sau.LOGNAME)

    cmd = [ EMERGE_PATH, '-q', '--sync' ]
    ret, out, err = sau.helpers.exec_cmd(cmd, timeout=3600)

    if ret != 0:
        log.warning("Portage sync failed:")
        for line in out.splitlines():
            log.warning("stdout: {}".format(line))
        for line in err.splitlines():
            log.warning("stderr: {}".format(line))

    if os.path.exists(EIX_UPDATE_PATH):
        cmd = [ EIX_UPDATE_PATH, '-q' ]
        ret, out, err = sau.helpers.exec_cmd(cmd, timeout=3600)

        if ret != 0:
            log.warning("eix-update failed:")
            for line in out.splitlines():
                log.warning("stdout: {}".format(line))
            for line in err.splitlines():
                log.warning("stderr: {}".format(line))


def pkg_upgrade():
    log = logging.getLogger(sau.LOGNAME)
    conf = sau.config

    _sync_portage()

    # [ebuild     U ] media-plugins/alsa-plugins-1.1.8 [1.1.6]
    pretend_re = re.compile(r'^\[ebuild ([^\]]*)\] ([^ ]+)( \[[^\]]+\])?')
    # media-plugins/alsa-plugins-1.1.8
    version_re = re.compile(r'^(.*/.*)-(\d+.*)$')

    ignore_re = re.compile(r'^(|.*caus.* rebuilds.*|.*scheduled for merge.*|.*waiting for lock on.*)$')
    
    default_version_sens = conf.getint('default', 'version_sensitivity', fallback=1)

    cmd = [ EMERGE_PATH, '--color', 'n', '-uDNpq', '@world' ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if not ret == 0:
        log.error('emerge pretend returned {}'.format(ret))
        for line in out.splitlines():
            log.error('stdout: {}'.format(line))
        for line in err.splitlines():
            log.error('stderr: {}'.format(line))
        return False

    do_rebuild = True
    for line in out.splitlines():
        if re.match(ignore_re, line):
            continue
        match = re.match(pretend_re, line)
        if not match:
            log.warning('No match on line: {}'.format(line))
            continue
        status = match.group(1)
        name = match.group(2)
        old = match.group(3)
        if not old:
            continue
        old = old.strip(' []')
        nmatch = re.match(version_re, name)
        name = nmatch.group(1)
        version = nmatch.group(2)

        sens = conf.getint('packages', name, fallback=default_version_sens)
        common = sau.helpers.version_diff(version, old)
        if sens <= common:
            log.info('{}-{} -> {} configured level {} <= pkg level {}'.format(name, old, version, sens, common))
        else:
            log.warning('{}-{} -> {} configured level {} > pkg level {}'.format(name, old, version, sens, common))
            do_rebuild = False

    if not do_rebuild:
        log.warning('Some packages require manual attention, did not upgrade')
        return False

    cmd = [ EMERGE_PATH, '--color', 'n', '-uDNq', '@world' ]
    ret, out, err = sau.helpers.exec_cmd(cmd, timeout=36000)

    if ret != 0 or err:
        log.warning('emerge returned {}'.format(ret))
        for line in out.splitlines():
            log.warning('stdout: {}'.format(line))
        for line in err.splitlines():
            log.warning('stderr: {}'.format(line))
    else:
        log.info('upgrade complete')
        for line in out.splitlines():
            if line.startswith(' * '):
                log.warning(line)

    cmd = [ EMERGE_PATH, '--color', 'n', '-q', '@preserved-rebuild' ]
    ret, out, err = sau.helpers.exec_cmd(cmd, timeout=36000)

    if ret != 0 or err:
        log.warning('preserved-rebuild returned {}'.format(ret))
        for line in out.splitlines():
            log.warning('stdout: {}'.format(line))
        for line in err.splitlines():
            log.warning('stderr: {}'.format(line))
    else:
        log.info('preserved-rebuild complete')
        for line in out.splitlines():
            if line.startswith(' * '):
                log.warning(line)

    if conf.getboolean('default', 'do_depclean', fallback=False):
        cmd = [ EMERGE_PATH, '--color', 'n', '-q', '--depclean' ]
        ret, out, err = sau.helpers.exec_cmd(cmd, timeout=3600)
        if ret != 0 or err:
            log.warning('depclean returned {}'.format(ret))
            for line in out.splitlines():
                log.warning('stdout: {}'.format(line))
            for line in err.splitlines():
                log.warning('stderr: {}'.format(line))
        else:
            log.info('depclean complete')
            for line in out.splitlines():
                if line.startswith(' * '):
                    log.warning(line)
